
import threading
import time
from os.path import join, dirname, abspath

import rclpy
from rclpy.node import Node,SetParametersResult,Parameter
from std_msgs.msg import Float64MultiArray,Int16MultiArray
from sensor_msgs.msg import Imu,MagneticField,NavSatStatus,NavSatFix
from geometry_msgs.msg import Vector3
from dataclasses import dataclass
from lifecycle_msgs.msg import State
from lifecycle_msgs.msg import Transition
from littleboss.lifecycle import LifecycleNode
from interfaces.msg import HapticSignal
from littleboss.sensors import PhysicalSensor, VirtualStreamSensor, DummySensor
from littleboss.utils import SENSOR_MAP
from typing import List
import gc
from rclpy.qos import QoSPresetProfiles
DATA_QOS = QoSPresetProfiles.SENSOR_DATA.value

ROOT_PATH = dirname(dirname(abspath(__file__)))
@dataclass
class Config:
    name_space: str
    rate: int
    type: str
    receive: bool
    send: bool

class LittleBoss(LifecycleNode):
    def __init__(self):
        super().__init__('littleboss_node')
        self.declare_parameters(
            namespace='',
            parameters=[(attr, None) for attr in Config.__annotations__]
        )
        self.logger = self.get_logger()  
        self.config = None
        self.sensor = None
        self.lock = threading.Lock()
        self.add_on_set_parameters_callback(self.config_update)
    
    def on_configure(self):
        self.logger.info("on_configure() is called")
        self.config = Config(
            **{attr: self.get_parameter(attr).value for attr in Config.__annotations__}
        ) if self.config is None else self.config
                
        try:
            with self.lock:
                self.start_sensor()
        except Exception as e:
            self.logger.error(f"LittleBoss init failed: {e}")
        return Transition.TRANSITION_CALLBACK_SUCCESS
            
    def on_cleanup(self):
        self.logger.info("on_cleanup() is called")
        self.stop_sensor()
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_activate(self): 
        self.logger.info("on_activate() is called")
        self.imu_pub = self.create_publisher(Float64MultiArray, f'{self.config.name_space}/imu', DATA_QOS)
        self.ori_pub = self.create_publisher(Float64MultiArray,  f'{self.config.name_space}/orientation', DATA_QOS)
        self.gps_pub = self.create_publisher(Float64MultiArray, f'{self.config.name_space}/gps', DATA_QOS)
        
        # self.imu_pub = self.create_publisher(Imu, f'{self.config.name_space}/imu', DATA_QOS)
        # self.ori_pub = self.create_publisher(MagneticField,  f'{self.config.name_space}/orientation', DATA_QOS)
        # self.gps_pub = self.create_publisher(NavSatFix, f'{self.config.name_space}/gps', DATA_QOS)
        # self.tof1_pub = self.create_publisher(Int16MultiArray, f'{self.config.name_space}/tof1', DATA_QOS)
        # self.tof2_pub = self.create_publisher(Int16MultiArray, f'{self.config.name_space}/tof2', DATA_QOS)
        
        self.hap_sub = self.create_subscription(HapticSignal, '/haptic_pattern', self.signal_haptics, 2)
        self.add_on_set_parameters_callback(self.config_update)
        rate = self.config.rate
        self.timer = self.create_timer(1/rate, self.publish_sensor_data)
        return Transition.TRANSITION_CALLBACK_SUCCESS

    def on_deactivate(self):
        self.logger.info("on_deactivate() is called")
        self.timer.cancel()
        # self.timer.destroy()
        self.timer = None
        
        # self.imu_pub.destroy()
        self.imu_pub = None
        # self.ori_pub.destroy()
        self.ori_pub = None
        # self.gps_pub.destroy()
        self.gps_pub = None
        self.tof1_pub = None
        self.tof2_pub = None
        
        # self.hap_sub.destroy()
        self.hap_sub = None
        gc.collect()
        return Transition.TRANSITION_CALLBACK_SUCCESS   

    def destroy_node(self) -> bool:
        self.stop_sensor()
        return super().destroy_node()
        
    def config_update(self, params: List[Parameter]):
        self.logger.info(f'Update Config: {[str(p.name) for p in params]}')
        for param in params:
            if param.name in self.config.__annotations__:
                self.config.__setattr__(param.name, param.value)

        return SetParametersResult(successful=True)
    
    def start_sensor(self):
        if not self.config.receive or not self.config.send: return  
        sensor_type = self.config.type
        if sensor_type.lower() not in ["physical", "virtual"]:
            raise Exception(f"Unknown sensor type: {sensor_type}")
        try:
            self.stop_sensor()
            if sensor_type.lower() == "physical":
                import serial
                self.sensor = PhysicalSensor(
                    port="/dev/ttyTHS2",
                    baudrate=115200,
                    bytesize=serial.EIGHTBITS,
                    parity=serial.PARITY_NONE,
                    stopbits=serial.STOPBITS_ONE,
                )
            elif sensor_type.lower() == "virtual":
                # self.sensor = VirtualStreamSensor(
                #     recv_port=9999,
                #     send_args=join(ROOT_PATH,'bluetooth.config')
                # )
                self.sensor = DummySensor()
        except Exception as e:
            raise Exception(f"Sensor init failed: {e}")
        
        self.sensor.reset()
        self.init_success = self.sensor.wait()
        if not self.init_success:
            raise Exception("Sensor init failed: timeout")
         
        self.logger.info("Sensor Init Success")
    
    def stop_sensor(self):
        if hasattr(self, "sensor") and self.sensor is not None:
            self.sensor.close()
            self.sensor = None
            # Logic to stop the camera
            self.logger.info("Little Boss stopped") 
            
    def publish_sensor_data(self):
        if self.sensor is None or not self.config.receive: return
        data = self.sensor.read()
        
        acc, gyro, ori, gps = self.break_down_data(data)
        self.imu_pub.publish(Float64MultiArray(data= acc + gyro))
        self.ori_pub.publish(Float64MultiArray(data=ori))
        self.gps_pub.publish(Float64MultiArray(data=gps))
        
        # imu, ori, gps, tof1, tof2 = self.break_down_data(data)
        # self.imu_pub.publish(Imu(
        #     linear_acceleration = Vector3(x=imu[0], y=imu[1], z=imu[2]),
        #     angular_velocity = Vector3(x=imu[3], y=imu[4], z=imu[5]),
        # ))
        
        # self.ori_pub.publish(MagneticField(
        #     magnetic_field=Vector3(x=ori[0], y=ori[1], z=ori[2]),
        # ))
        
        # self.gps_pub.publish(NavSatFix(
        #     status = NavSatStatus(
        #         status=gps[0],
        #         service=gps[1]
        #     ),
        #     latitude=gps[2],
        #     longitude=gps[3],
        #     altitude=gps[4],
        #     position_covariance=[
        #         gps[8], gps[9], gps[10],
        #         gps[9], gps[11], gps[12],
        #         gps[10], gps[12], gps[13],
        #     ],
        #     position_covariance_type=NavSatFix.COVARIANCE_TYPE_KNOWN
        # ))        
        # self.tof1_pub.publish(Int16MultiArray(data=tof1))        
        # self.tof2_pub.publish(Int16MultiArray(data=tof2))

    def signal_haptics(self, signal:HapticSignal):
        if self.sensor is None or not self.config.send: return
        # Decode the message
        self.logger.info(f"Writing {signal.signal}")
        self.set_intensity(signal.intensity)
        if signal.delay > 0:
            time.sleep(signal.delay)
        self.sensor.write(signal.signal.encode())
                
    def set_intensity(self, intensity):
        # TODO: Wait for the hardware implementation
        pass
    
    @staticmethod
    def break_down_data(data):
        try:
            curr_acc = [
                data[SENSOR_MAP["ax"]],
                data[SENSOR_MAP["ay"]],
                data[SENSOR_MAP["az"]],
            ]
            curr_gyro = [
                data[SENSOR_MAP["gx"]],
                data[SENSOR_MAP["gy"]],
                data[SENSOR_MAP["gz"]],
            ]
            curr_orient = [
                data[SENSOR_MAP["roll"]],
                data[SENSOR_MAP["pitch"]],
                data[SENSOR_MAP["yaw"]],
            ]
            curr_coord = [data[SENSOR_MAP["lat"]], data[SENSOR_MAP["lng"]]]
            return curr_acc, curr_gyro, curr_orient, curr_coord
        except RuntimeError:
            print("break_down_data failed: Likely that data list is empty.")
        return (
            [0.0 for _ in range(3)],
            [0.0 for _ in range(3)],
            [0.0 for _ in range(3)],
            [0.0 for _ in range(2)],
        )


def main(args=None):
    rclpy.init(args=args)
    node = LittleBoss()
    try:
        rclpy.spin(node)
    except KeyboardInterrupt:
        if node.state != State.PRIMARY_STATE_FINALIZED:
            node.stop_sensor()
    rclpy.shutdown()
    
if __name__ == '__main__':
    main()
