import struct
import socket
import ast
import time
import json
import random
try:
    import serial
    import bluetooth
except:
    pass
from littleboss.utils import SENSOR_FOMAT,_unpack,g_unpack
# build an abstract class for receiving and sending data from sensor and IMU alike
## then implement the child classes for each sensor, virtual or physical

# the abstract class
class SensorInterface:
    def __init__(self):
        self._data = [0.0 for _ in range(11)]

    def get_data(self):
        return self._data

    def set_data(self, data):
        self._data = data

    def read(self):
        raise NotImplementedError

    def write(self):
        raise NotImplementedError
    
    def reset(self):
        raise NotImplementedError
    
    def close(self):
        raise NotImplementedError
    
    def __str__(self):  
        reformat =[]
        for i,d in enumerate(self._data):
            if i < 9:
                reformat.append(f'{d :<8.3f}' if d<0 else f' {d :<7.3f}')
            else:
                reformat.append(f'{d :<10.6f}')
        return f'[{", ".join(reformat)}]'
        
    def __repr__(self):
        return str(self)
    
    def __del__(self):
        pass
        
        
# the child class for the IMU
class PhysicalSensor(SensorInterface):
    def __init__(self, port, baudrate, **kwargs):
        super().__init__()
        self._port = port
        self._baudrate = baudrate
        self._serial = serial.Serial(self._port, self._baudrate, **kwargs)
    
    def write(self, data):
        self._serial.write(data)
        
    def read(self):    
        try:
            # read 14 bytes from the serial
            # map the header type to the actual sensor data 
            header_type={1:'a',2:'g',3:'m', 4:'c'}
            return_data={}
            # while return data does not contain a,g,m,c
            for _ in range(4):
                r = self._serial.read(14)
                data, header = self.unpack_imu(r)
                return_data[header_type[header]] = data
            if len(return_data['a']) != 3 or len(return_data['g']) != 3 or len(return_data['m']) != 3 or len(return_data['c']) != 2:
                raise Exception('Invalid data')
            self._data = return_data['a'] + return_data['g'] + return_data['m'] + return_data['c']
        except:
            self.reset()
            #self.wait()
            self._data = [0.0 for _ in range(11)]
        return self.get_data()
    
    def close(self):
        self._serial.close()        
    
    def wait(self):    
        # wait for the serial port to be ready
        wait_sec= 0 
        while self._serial.inWaiting() <= 0:
            time.sleep(1)
            wait_sec+=1
            if wait_sec > 10:
                return False
        return True
        
    def reset(self):
        self._serial.flushInput()
        self._serial.flushOutput()
    
    @staticmethod
    def unpack_imu(data):
        header = data[0]
        length = data[1]
        data = data[2:]
        if len(data) < 4:
            return [0.0 for _ in range(3)]
        r = [struct.unpack('f',data[i:i+4])[0] for i in range(0, length, 4)]
        return r,header
            

    
class VirtualStreamSensor(SensorInterface):
    def __init__(self, recv_port, send_args=None):
        super().__init__()
        # create a socket object
        self.recv_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM) 
        self.recv_host = socket.gethostname()   
        self.recv_port = recv_port
                
        if send_args is None:
            return
        send_args = json.load(open(send_args, 'r'))
        # Create the client socket
        self.send_socket = bluetooth.BluetoothSocket(bluetooth.RFCOMM)
        self.send_host = send_args["host"]
        self.send_port = send_args["port"]
            
    def write(self, data):
        if self.send_socket is not None:
            self.send_socket.send(data)
        
    def read(self):    
        try:
            self.recv_socket.send(b'get data')
            msg = self.recv_socket.recv(1024)      
            data = self.parse_line(msg.decode('ascii'))
            for i in range(6):
                self._data[i] = data[i]
        except:
            pass
        return self.get_data()
        
    def close(self):
        self.recv_socket.send(b'quit')
        try:
            self.recv_socket.close()
            self.send_socket.close()
        except:
            pass
    
    def wait(self):   
        # connection to hostname on the port. 
        try:
            self.recv_socket.connect((self.recv_host, self.recv_port))
            self.send_socket.connect((self.send_host, self.send_port))
            return True              
        except:
            return False
        
    def reset(self):
        pass
    @staticmethod
    def parse_line(line):
        return ast.literal_eval(line)
    

    
class DummySensor(SensorInterface):
    def __init__(self):
        super().__init__()
        self.count=0
            
    def write(self, data):
        print(f'Dummy Sensor Writing {data}')
        
    def read(self):    
        if self.count % 100 == 0:
            print('Dummy Sensor Reading')
        self.count+=1   
        self._data = [random.random() for _ in range(11)]
        return self.get_data()
        
    def close(self):
        print('Dummy Sensor Closing')
    
    def wait(self):   
        return True              
        
    def reset(self):
        pass
    
class RDKSensor:
    def __init__(self, port, baudrate, **kwargs):
        self._port = port
        self._baudrate = baudrate
        try:
            self._serial = serial.Serial(self._port, self._baudrate, **kwargs)
        except:
            pass
        self.buffer = b''
        self.data = dict(
            imu = [0.0 for _ in range(6)],
            mag = [0.0 for _ in range(3)],
            gps = [0.0 for _ in range(14)],
            tof1 = [0.0 for _ in range(16)],
            tof2 = [0.0 for _ in range(16)]
        )
    def write(self, data):
        self._serial.write(data)
        
    def read(self):    
        self.buffer += self._serial.read(self._serial.inWaiting())
    
    def close(self):
        self._serial.close()        
    
    def wait(self):    
        # wait for the serial port to be ready
        wait_sec= 0 
        while self._serial.inWaiting() <= 0:
            time.sleep(1)
            wait_sec+=1
            if wait_sec > 10:
                return False
        return True
        
    def reset(self):
        self._serial.flushInput()
        self._serial.flushOutput()
    
    def unpack(self):
        data = self.buffer
        while len(data) > 0:
            header = _unpack(data[0:1], 'B')
            checksum = _unpack( data[1:2], 'B')
            length = _unpack( data[2:3], 'B')
            payload = data[ 3: 3 + length]
            data = data[3 + length:]
            
            if header not in SENSOR_FOMAT:
                raise ValueError(f'Header {header} is not supported')        

            sensor_info = SENSOR_FOMAT[header]
            sensor_type= sensor_info['type']
            
            if sensor_type == 'gps':
                status = _unpack(payload[0:1], 'B')
                service = _unpack(payload[1:2], 'B')
                length -= 2
                payload = payload[2:]
            r = g_unpack(payload, sensor_info, checksum, length)
            
            if sensor_type == 'gps':
                # print(f'status: {status}, service: {service}')
                r= [status, service] + r
            print(f'{sensor_type}: {r}')
            self.data[sensor_type] = r
            